#
##
##  This file is part of pyFormex 1.0.7  (Mon Jun 17 12:20:39 CEST 2019)
##  pyFormex is a tool for generating, manipulating and transforming 3D
##  geometrical models by sequences of mathematical operations.
##  Home page: http://pyformex.org
##  Project page:  http://savannah.nongnu.org/projects/pyformex/
##  Copyright 2004-2019 (C) Benedict Verhegghe (benedict.verhegghe@ugent.be)
##  Distributed under the GNU General Public License version 3 or later.
##
##  This program is free software: you can redistribute it and/or modify
##  it under the terms of the GNU General Public License as published by
##  the Free Software Foundation, either version 3 of the License, or
##  (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##  GNU General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see http://www.gnu.org/licenses/.
##
"""SaveLoad

Example illustrating the saving to and loading from a file in
the new pyFormex zip format (PZF).

"""
from __future__ import absolute_import, division, print_function

_status = 'checked'
_level = 'normal'
_topics = ['save']
_techniques = ['pzf', 'field', 'savePZF', 'loadPZF']

from pyformex.gui.draw import *
from pyformex.plugins.saveload import *
from pyformex.simple import cylinder
from pyformex.mesh import rectangleWithHole
from pyformex.examples.Spirals import createSpiral


def run():
    clear()
    smoothwire()
    view('xz')

    F = cylinder(3, 4, 12, 2)
    M = rectangleWithHole(10, 10, 3, 6, 3)
    M += M.reflect(0)
    M += M.reflect(1)
    T = TriSurface.read(getcfg('datadir') / 'man.off')
    T = T.scale(4).align('00-', [0., 0., 0.])
    PL = createSpiral(nseg=100, turns=1.8, phi=30., alpha=70., c=0.8)
    PL.attrib(linewidth=3)
    F.setProp('range')
    M.attrib(color='fld:dist3n', lighting=False)
    T.attrib(color='gold', mode='smooth')
    data = length(M.coords - [5., 3., 0.])
    M.addField('node', data, 'dist')
    data = abs(M.coords - [5., 3., 0.])
    data /= data.max()
    M.addField('node', data, 'dist3n')
    data = abs(M.centroids() - [5., 3., 0.])
    data /= data.max()
    M.addField('elemc', data, 'dist3c')
    X = Coords('263374').scale(5).rotate(-90., 0).trl([10., 10., 10.])
    X[-1].y -= 5
    C = BezierSpline(X, degree=3, closed=True)
    C.attrib(linewidth=2, color=pyformex_pink)
    draw([F, M, T, PL, C, X])

    CS = CoordSys()
    drawAxes(CS, size=0.5, psize=0.0)

    pause(msg="""..

SaveLoad example
----------------
First we have created some geometry:

F: Formex: a cylinder of rectangular panels colored by its prop='range'
M: Mesh: a square plate with circular hole, colored from a Field being the
   3D-distance from a point (the center of the black zone). This object is
   drawn without lighting (set in the attrib dict).
T: TriSurface: a surface representing a man, colored in gold, standing
   inside the cylinder.

You can now change the camera settings like you wish.
When you hit the button, we will ask you for a .pzf filename to save the
model to.
""")

    fn = askNewFilename(getcfg('workdir'), 'pzf')
    if not fn:
        return

    savePZF(fn, F=F, M=M, T=T, spiral=PL, CS=CS, curve=C, X=X, _camera=True)
    clear()
    F = M = T = PL = CS = C = X = None
    print("Archive contents:")
    print(listPZF(fn))
    pause(msg="""
Now we have saved the whole model in PZF format and
erased it from the screen and from memory.
When you hit the PLAY button, we will reload the model
from the saved file, and display it again.
""")

    d = loadPZF(fn)
    print("Loaded %s objects" % len(d))
    drawn = draw(d.values())
    print("Drawn %s objects" % len(drawn))
    # currently, CoordSys can not be drawn directly
    drawAxes(d['CS'], size=0.5, psize=0.0)
    # Set the camera as stored in the file
    if '_camera' in d:
        pf.canvas.camera.apply(d['_camera'])
        pf.canvas.update()

    print(listPZFobj(fn))

if __name__ == '__draw__':
    run()

# End
